==> v0.23.2.tar.gz: libretube/app/src/main/java/com/github/libretube/ui/fragments/ChannelFragment.kt <==
package com.github.libretube.ui.fragments

import android.content.res.Configuration
import android.os.Bundle
import android.os.Parcelable
import android.util.Log
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import androidx.core.os.bundleOf
import androidx.core.view.children
import androidx.core.view.isGone
import androidx.core.view.isVisible
import androidx.lifecycle.lifecycleScope
import androidx.navigation.fragment.navArgs
import androidx.recyclerview.widget.GridLayoutManager
import androidx.recyclerview.widget.RecyclerView
import com.github.libretube.R
import com.github.libretube.api.RetrofitInstance
import com.github.libretube.api.obj.ChannelTab
import com.github.libretube.constants.IntentData
import com.github.libretube.databinding.FragmentChannelBinding
import com.github.libretube.enums.ShareObjectType
import com.github.libretube.extensions.TAG
import com.github.libretube.extensions.ceilHalf
import com.github.libretube.extensions.formatShort
import com.github.libretube.extensions.toID
import com.github.libretube.helpers.ImageHelper
import com.github.libretube.helpers.NavigationHelper
import com.github.libretube.obj.ChannelTabs
import com.github.libretube.obj.ShareData
import com.github.libretube.ui.adapters.SearchChannelAdapter
import com.github.libretube.ui.adapters.VideosAdapter
import com.github.libretube.ui.base.DynamicLayoutManagerFragment
import com.github.libretube.ui.dialogs.ShareDialog
import com.github.libretube.ui.extensions.addOnBottomReachedListener
import com.github.libretube.ui.extensions.setupSubscriptionButton
import com.github.libretube.ui.sheets.AddChannelToGroupSheet
import com.github.libretube.util.deArrow
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import retrofit2.HttpException
import java.io.IOException

class ChannelFragment : DynamicLayoutManagerFragment() {
    private var _binding: FragmentChannelBinding? = null
    private val binding get() = _binding!!
    private val args by navArgs<ChannelFragmentArgs>()

    private var channelId: String? = null
    private var channelName: String? = null
    private var channelAdapter: VideosAdapter? = null
    private var isLoading = true

    private val possibleTabs = arrayOf(
        ChannelTabs.Shorts,
        ChannelTabs.Livestreams,
        ChannelTabs.Playlists,
        ChannelTabs.Albums
    )
    private var channelTabs: List<ChannelTab> = emptyList()
    private var nextPages = Array<String?>(5) { null }
    private var searchChannelAdapter: SearchChannelAdapter? = null

    private var isAppBarFullyExpanded: Boolean = true
    private var recyclerViewState: Parcelable? = null

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        channelName = args.channelName
            ?.replace("/c/", "")
            ?.replace("/user/", "")
        channelId = args.channelId
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {
        _binding = FragmentChannelBinding.inflate(inflater, container, false)
        return binding.root
    }

    override fun setLayoutManagers(gridItems: Int) {
        _binding?.channelRecView?.layoutManager = GridLayoutManager(
            context,
            gridItems.ceilHalf()
        )
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)

        // Check if the AppBarLayout is fully expanded
        binding.channelAppBar.addOnOffsetChangedListener { _, verticalOffset ->
            isAppBarFullyExpanded = verticalOffset == 0
        }

        // Determine if the child can scroll up
        binding.channelRefresh.setOnChildScrollUpCallback { _, _ ->
            !isAppBarFullyExpanded
        }

        binding.channelRefresh.setOnRefreshListener {
            fetchChannel()
        }

        binding.channelRecView.addOnBottomReachedListener {
            if (_binding == null || isLoading) return@addOnBottomReachedListener

            loadNextPage()
        }

        binding.channelRecView.addOnScrollListener(object : RecyclerView.OnScrollListener() {
            override fun onScrollStateChanged(recyclerView: RecyclerView, newState: Int) {
                super.onScrollStateChanged(recyclerView, newState)
                recyclerViewState = binding.channelRecView.layoutManager?.onSaveInstanceState()
            }
        })

        fetchChannel()
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private fun loadNextPage() = lifecycleScope.launch {
        val binding = _binding ?: return@launch

        binding.channelRefresh.isRefreshing = true
        isLoading = true

        try {
            if (binding.tabChips.checkedChipId == binding.videos.id) {
                fetchChannelNextPage(nextPages[0] ?: return@launch).let {
                    nextPages[0] = it
                }
            } else {
                val currentTabIndex = binding.tabChips.children.indexOfFirst {
                    it.id == binding.tabChips.checkedChipId
                }
                val channelTab = channelTabs.first { tab ->
                    tab.name == possibleTabs[currentTabIndex - 1].identifierName
                }
                val nextPage = nextPages[currentTabIndex] ?: return@launch
                fetchTabNextPage(nextPage, channelTab).let {
                    nextPages[currentTabIndex] = it
                }
            }
        } catch (e: Exception) {
            Log.e("error fetching tabs", e.toString())
        }
    }.invokeOnCompletion {
        _binding?.channelRefresh?.isRefreshing = false
        isLoading = false
    }

    private fun fetchChannel() = lifecycleScope.launch {
        isLoading = true
        binding.channelRefresh.isRefreshing = true

        val response = try {
            withContext(Dispatchers.IO) {
                if (channelId != null) {
                    RetrofitInstance.api.getChannel(channelId!!)
                } else {
                    RetrofitInstance.api.getChannelByName(channelName!!)
                }.apply {
                    relatedStreams = relatedStreams.deArrow()
                }
            }
        } catch (e: IOException) {
            Log.e(TAG(), "IOException, you might not have internet connection")
            return@launch
        } catch (e: HttpException) {
            Log.e(TAG(), "HttpException, unexpected response")
            return@launch
        } finally {
            _binding?.channelRefresh?.isRefreshing = false
            isLoading = false
        }
        val binding = _binding ?: return@launch

        // needed if the channel gets loaded by the ID
        channelId = response.id
        channelName = response.name
        val shareData = ShareData(currentChannel = response.name)

        val channelId = channelId ?: return@launch

        binding.channelSubscribe.setupSubscriptionButton(
            channelId,
            channelName,
            binding.notificationBell
        )

        binding.channelSubscribe.setOnLongClickListener {
            AddChannelToGroupSheet().apply {
                arguments = bundleOf(IntentData.channelId to channelId)
            }.show(childFragmentManager)

            true
        }

        binding.channelShare.setOnClickListener {
            val bundle = bundleOf(
                IntentData.id to channelId.toID(),
                IntentData.shareObjectType to ShareObjectType.CHANNEL,
                IntentData.shareData to shareData
            )
            val newShareDialog = ShareDialog()
            newShareDialog.arguments = bundle
            newShareDialog.show(childFragmentManager, ShareDialog::class.java.name)
        }

        nextPages[0] = response.nextpage
        isLoading = false
        binding.channelRefresh.isRefreshing = false

        binding.channelCoordinator.isVisible = true

        binding.channelName.text = response.name
        if (response.verified) {
            binding.channelName
                .setCompoundDrawablesWithIntrinsicBounds(0, 0, R.drawable.ic_verified, 0)
        }
        binding.channelSubs.text = resources.getString(
            R.string.subscribers,
            response.subscriberCount.formatShort()
        )
        if (response.description.orEmpty().isBlank()) {
            binding.channelDescription.isGone = true
        } else {
            binding.channelDescription.text = response.description.orEmpty().trim()
        }

        ImageHelper.loadImage(response.bannerUrl, binding.channelBanner)
        ImageHelper.loadImage(response.avatarUrl, binding.channelImage, true)

        binding.channelImage.setOnClickListener {
            NavigationHelper.openImagePreview(
                requireContext(),
                response.avatarUrl ?: return@setOnClickListener
            )
        }

        binding.channelBanner.setOnClickListener {
            NavigationHelper.openImagePreview(
                requireContext(),
                response.bannerUrl ?: return@setOnClickListener
            )
        }

        // recyclerview of the videos by the channel
        channelAdapter = VideosAdapter(
            response.relatedStreams.toMutableList(),
            forceMode = VideosAdapter.Companion.LayoutMode.CHANNEL_ROW
        )
        binding.channelRecView.adapter = channelAdapter

        setupTabs(response.tabs)
    }

    private fun setupTabs(tabs: List<ChannelTab>) {
        this.channelTabs = tabs

        val binding = _binding ?: return

        binding.tabChips.children.forEach { chip ->
            val resourceTab = possibleTabs.firstOrNull { it.chipId == chip.id }
            resourceTab?.let { resTab ->
                if (tabs.any { it.name == resTab.identifierName }) chip.isVisible = true
            }
        }

        binding.tabChips.setOnCheckedStateChangeListener { _, _ ->
            when (binding.tabChips.checkedChipId) {
                binding.videos.id -> {
                    binding.channelRecView.adapter = channelAdapter
                }

                else -> {
                    possibleTabs.first { binding.tabChips.checkedChipId == it.chipId }.let {
                        val tab = tabs.first { tab -> tab.name == it.identifierName }
                        loadChannelTab(tab)
                    }
                }
            }
        }

        // Load selected chip content if it's not videos tab.
        possibleTabs.firstOrNull { binding.tabChips.checkedChipId == it.chipId }?.let {
            val tab = tabs.first { tab -> tab.name == it.identifierName }
            loadChannelTab(tab)
        }
    }

    private fun loadChannelTab(tab: ChannelTab) = lifecycleScope.launch {
        binding.channelRefresh.isRefreshing = true
        isLoading = true

        val response = try {
            withContext(Dispatchers.IO) {
                RetrofitInstance.api.getChannelTab(tab.data)
            }.apply {
                content = content.deArrow()
            }
        } catch (e: Exception) {
            return@launch
        }
        nextPages[channelTabs.indexOf(tab) + 1] = response.nextpage

        val binding = _binding ?: return@launch

        searchChannelAdapter = SearchChannelAdapter()
        binding.channelRecView.adapter = searchChannelAdapter
        searchChannelAdapter?.submitList(response.content)

        binding.channelRefresh.isRefreshing = false
        isLoading = false
    }

    private suspend fun fetchChannelNextPage(nextPage: String): String? {
        val response = withContext(Dispatchers.IO) {
            RetrofitInstance.api.getChannelNextPage(channelId!!, nextPage).apply {
                relatedStreams = relatedStreams.deArrow()
            }
        }

        channelAdapter?.insertItems(response.relatedStreams)

        return response.nextpage
    }

    private suspend fun fetchTabNextPage(nextPage: String, tab: ChannelTab): String? {
        val newContent = withContext(Dispatchers.IO) {
            RetrofitInstance.api.getChannelTab(tab.data, nextPage)
        }.apply {
            content = content.deArrow()
        }

        searchChannelAdapter?.let {
            it.submitList(it.currentList + newContent.content)
        }

        return newContent.nextpage
    }


    override fun onConfigurationChanged(newConfig: Configuration) {
        super.onConfigurationChanged(newConfig)
        // manually restore the recyclerview state due to https://github.com/material-components/material-components-android/issues/3473
        binding.channelRecView.layoutManager?.onRestoreInstanceState(recyclerViewState)
    }
}

==> v0.25.2.tar.gz: libretube/app/src/main/java/com/github/libretube/ui/fragments/ChannelFragment.kt <==
package com.github.libretube.ui.fragments

import android.os.Bundle
import android.util.Log
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import androidx.core.os.bundleOf
import androidx.core.view.isGone
import androidx.core.view.isVisible
import androidx.fragment.app.Fragment
import androidx.lifecycle.lifecycleScope
import androidx.navigation.fragment.navArgs
import androidx.recyclerview.widget.RecyclerView
import androidx.viewpager2.adapter.FragmentStateAdapter
import androidx.viewpager2.widget.ViewPager2
import com.github.libretube.R
import com.github.libretube.api.RetrofitInstance
import com.github.libretube.api.obj.ChannelTab
import com.github.libretube.api.obj.StreamItem
import com.github.libretube.constants.IntentData
import com.github.libretube.databinding.FragmentChannelBinding
import com.github.libretube.enums.ShareObjectType
import com.github.libretube.extensions.TAG
import com.github.libretube.extensions.formatShort
import com.github.libretube.extensions.toID
import com.github.libretube.helpers.ClipboardHelper
import com.github.libretube.helpers.ImageHelper
import com.github.libretube.helpers.NavigationHelper
import com.github.libretube.obj.ShareData
import com.github.libretube.ui.adapters.VideosAdapter
import com.github.libretube.ui.base.DynamicLayoutManagerFragment
import com.github.libretube.ui.dialogs.ShareDialog
import com.github.libretube.ui.extensions.setupSubscriptionButton
import com.github.libretube.ui.sheets.AddChannelToGroupSheet
import com.github.libretube.util.deArrow
import com.google.android.material.tabs.TabLayoutMediator
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import retrofit2.HttpException
import java.io.IOException

class ChannelFragment : DynamicLayoutManagerFragment() {
    private var _binding: FragmentChannelBinding? = null
    private val binding get() = _binding!!
    private val args by navArgs<ChannelFragmentArgs>()

    private var channelId: String? = null
    private var channelName: String? = null
    private var channelAdapter: VideosAdapter? = null
    private var isLoading = true

    private lateinit var channelContentAdapter: ChannelContentAdapter

    private var nextPages = Array<String?>(5) { null }
    private var isAppBarFullyExpanded: Boolean = true
    private val tabList = mutableListOf<ChannelTab>()

    private val tabNamesMap = mapOf(
        VIDEOS_TAB_KEY to R.string.videos,
        "shorts" to R.string.yt_shorts,
        "livestreams" to R.string.livestreams,
        "playlists" to R.string.playlists,
        "albums" to R.string.albums
    )

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        channelName = args.channelName
            ?.replace("/c/", "")
            ?.replace("/user/", "")
        channelId = args.channelId
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {
        _binding = FragmentChannelBinding.inflate(inflater, container, false)
        return binding.root
    }

    override fun setLayoutManagers(gridItems: Int) {}

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        // Check if the AppBarLayout is fully expanded
        binding.channelAppBar.addOnOffsetChangedListener { _, verticalOffset ->
            isAppBarFullyExpanded = verticalOffset == 0
        }

        binding.pager.reduceDragSensitivity()

        // Determine if the child can scroll up
        binding.channelRefresh.setOnChildScrollUpCallback { _, _ ->
            !isAppBarFullyExpanded
        }

        binding.channelRefresh.setOnRefreshListener {
            fetchChannel()
        }

        fetchChannel()
    }

    // adjust sensitivity due to the issue of viewpager2 with SwipeToRefresh https://issuetracker.google.com/issues/138314213
    private fun ViewPager2.reduceDragSensitivity() {
        val recyclerViewField = ViewPager2::class.java.getDeclaredField("mRecyclerView")
        recyclerViewField.isAccessible = true
        val recyclerView = recyclerViewField.get(this) as RecyclerView

        val touchSlopField = RecyclerView::class.java.getDeclaredField("mTouchSlop")
        touchSlopField.isAccessible = true
        val touchSlop = touchSlopField.get(recyclerView) as Int
        touchSlopField.set(recyclerView, touchSlop * 3)
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private fun fetchChannel() = lifecycleScope.launch {
        isLoading = true
        _binding?.channelRefresh?.isRefreshing = true

        val response = try {
            withContext(Dispatchers.IO) {
                if (channelId != null) {
                    RetrofitInstance.api.getChannel(channelId!!)
                } else {
                    RetrofitInstance.api.getChannelByName(channelName!!)
                }.apply {
                    relatedStreams = relatedStreams.deArrow()
                }
            }
        } catch (e: IOException) {
            Log.e(TAG(), "IOException, you might not have internet connection")
            return@launch
        } catch (e: HttpException) {
            Log.e(TAG(), "HttpException, unexpected response")
            return@launch
        } finally {
            _binding?.channelRefresh?.isRefreshing = false
            isLoading = false
        }
        val binding = _binding ?: return@launch

        // needed if the channel gets loaded by the ID
        channelId = response.id
        channelName = response.name
        val shareData = ShareData(currentChannel = response.name)

        val channelId = channelId ?: return@launch

        binding.channelSubscribe.setupSubscriptionButton(
            channelId,
            channelName,
            binding.notificationBell
        ) { isSubscribed ->
            _binding?.addToGroup?.isVisible = isSubscribed
        }

        binding.channelShare.setOnClickListener {
            val bundle = bundleOf(
                IntentData.id to channelId.toID(),
                IntentData.shareObjectType to ShareObjectType.CHANNEL,
                IntentData.shareData to shareData
            )
            val newShareDialog = ShareDialog()
            newShareDialog.arguments = bundle
            newShareDialog.show(childFragmentManager, ShareDialog::class.java.name)
        }

        binding.addToGroup.setOnClickListener {
            AddChannelToGroupSheet().apply {
                arguments = bundleOf(IntentData.channelId to channelId)
            }.show(childFragmentManager)
        }

        binding.playAll.setOnClickListener {
            val firstVideoId =
                response.relatedStreams.firstOrNull()?.url?.toID() ?: return@setOnClickListener

            NavigationHelper.navigateVideo(requireContext(), firstVideoId, channelId = channelId)
        }

        nextPages[0] = response.nextpage
        isLoading = false
        binding.channelRefresh.isRefreshing = false

        binding.channelCoordinator.isVisible = true

        binding.channelName.text = response.name
        binding.channelName.setOnLongClickListener {
            ClipboardHelper.save(requireContext(), text = response.name.orEmpty())
            true
        }

        if (response.verified) {
            binding.channelName
                .setCompoundDrawablesWithIntrinsicBounds(0, 0, R.drawable.ic_verified, 0)
        }
        binding.channelSubs.text = resources.getString(
            R.string.subscribers,
            response.subscriberCount.formatShort()
        )
        if (response.description.orEmpty().isBlank()) {
            binding.channelDescription.isGone = true
        } else {
            binding.channelDescription.text = response.description.orEmpty().trim()
        }

        ImageHelper.loadImage(response.bannerUrl, binding.channelBanner)
        ImageHelper.loadImage(response.avatarUrl, binding.channelImage, true)

        binding.channelImage.setOnClickListener {
            NavigationHelper.openImagePreview(
                requireContext(),
                response.avatarUrl ?: return@setOnClickListener
            )
        }

        binding.channelBanner.setOnClickListener {
            NavigationHelper.openImagePreview(
                requireContext(),
                response.bannerUrl ?: return@setOnClickListener
            )
        }

        channelContentAdapter = ChannelContentAdapter(
            tabList,
            response.relatedStreams,
            response.nextpage,
            channelId,
            this@ChannelFragment
        )
        binding.pager.adapter = channelContentAdapter
        TabLayoutMediator(binding.tabParent, binding.pager) { tab, position ->
            tab.text = tabList[position].name
        }.attach()

        channelAdapter = VideosAdapter(
            response.relatedStreams.toMutableList(),
            forceMode = VideosAdapter.Companion.LayoutMode.CHANNEL_ROW
        )
        tabList.clear()

        val tabs = listOf(ChannelTab(VIDEOS_TAB_KEY, "")) + response.tabs
        for (channelTab in tabs) {
            val tabName = tabNamesMap[channelTab.name]?.let { getString(it) }
                ?: channelTab.name.replaceFirstChar(Char::titlecase)
            tabList.add(ChannelTab(tabName, channelTab.data))
        }
        channelContentAdapter.notifyItemRangeChanged(0, tabList.size - 1)
    }

    companion object {
        private const val VIDEOS_TAB_KEY = "videos"
    }
}

class ChannelContentAdapter(
    private val list: List<ChannelTab>,
    private val videos: List<StreamItem>,
    private val nextPage: String?,
    private val channelId: String?,
    fragment: Fragment
) : FragmentStateAdapter(fragment) {
    override fun getItemCount() = list.size

    override fun createFragment(position: Int) = ChannelContentFragment().apply {
        arguments = bundleOf(
            IntentData.tabData to list[position],
            IntentData.videoList to videos.toMutableList(),
            IntentData.channelId to channelId,
            IntentData.nextPage to nextPage
        )
    }
}
